// (C) 2009 THQ Inc. Not for commercial use or redistribution without permission. All rights reserved.

#pragma once

#ifndef ASTAR_H
#define ASTAR_H



#include "RegionPool.h"

#include <vector>
#include <algorithm>
#include <assert.h>



//
// Inputs for this template:
//
// - UserType: the type of the nodes in the AStar routine (like Math::Point2i)
//
// - Environment: class that gives the AStar the information it needs to do it's job. Needs the following functions:
//
//		// Returns the heuristic value of the current point to the goal
//		CostType	Heuristic( const UserType &current, const UserType &goal, CostType gval ) const;	
//	
//		// Adds the neighbours of the given 'point' node to the neighbours vector. It also add the additional g-cost for a move to that neighbour to the gcost vector
//		void		GetNeighbours( const UserType &parent_point, const UserType &point, std::vector< UserType > &neighbours, std::vector< CostType > &gcost ) const;
//
//		// Returns true if the exit condition for this AStar search was met
//		bool		TestExitCondition( const UserType &current_point, const UserType &goal ) const;
//
//		// Returns true if the cell is valid as a partial result
//		bool		ValidatePartialResult( const UserTYpe& current_point ) const;
//
// - TouchedList: class that handles the open/closed list of AStar. Needs the following functions:
//
//     Adds a node to the list. May be open or closed.
//     void Push( const UserType& userType, void* astarNode, bool bOpen );
//
//     Updates a node currently in the list, changing it's open/closed state.
//     void Update( iterator it, bool bOpen );
//
//     Finds a node in the list. If not found, this function returns End().
//     iterator Find( const UserType& userType );
//
//     Return the beginning and end of the list.
//     const_iterator Begin() const;
//     const_iterator End() const;
//
//     Returns the astar node of a specific list position.
//     void* GetNode( const_iterator it ) const;
// 
//     Clears the list for a new search.
//     void StartNewSearch();


enum AStarSearchResult
{
	AStarNoPath,
	AStarGotPathToGoal,
	AStarGotPathImperfect,
};

template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
class AStar
{
public:

	// mempoolhandle: memory pool handle. Make sure that it is a SEPARATE memory pool, as this class calls MemPoolClear()
	AStar( TouchedList& touchedList, RegionPool& regionPool )
	:	m_touchedList( touchedList ),
		m_nodePool( regionPool ),
		m_goal( 0 ),
		m_searchOpen( false ),
		m_nodesVisited( 0 ),
		m_nodesClosed( 0 ),
		m_lowestHNodeSoFar( 0 )
	{
		m_currentSolutionIt = m_solution.rend();
		m_startVector.reserve(8);
	};

	~AStar()
	{	
		FreeAllNodes();
	};

	//! Main pathfinding function. This function searches from the start to the goal, and returns the result
	AStarSearchResult		DoFindPath( const UserType &start, const UserType &goal, const Environment &env, unsigned long maxNodes );

	//! Main pathfinding function. This function searches from the start to the goal, and returns the result
	AStarSearchResult		DoFindOpen( UserType &firstOpen, const UserType &start, const UserType &goal, const Environment &env, unsigned long maxNodes );

	//! Main pathfinding function. This function searches from all start positions to the goal, and returns the result
	AStarSearchResult		DoFindPath( const UserType* startsBegin, const UserType* startsEnd, const UserType &goal, const Environment &env, unsigned long maxNodes );

	// Functions for traversing the solution:
	UserType *				GetSolutionStart();			// Returns the start of the solution, or NULL if there is no solution
	UserType *				GetSolutionNext();			// Returns the next node in the solution, or NULL if there is no solution or no next point
	void					FreeAllNodes();				// Should be called after the solution is no longer needed.


	// Debug values:
	unsigned int			GetNodesVisited() const		{ return m_nodesVisited; }
	unsigned int			GetNodesClosed() const		{ return m_nodesClosed; }
	TouchedList&			GetTouchedListDEBUG()		{ return m_touchedList; }


private:

	// Sets the start and the goal of the search
	void					SetStartAndGoal( const UserType *startVecBegin, const UserType *startVecEnd, const UserType &goal, const Environment &env );

	enum State
	{
		Searching,
		Succeeded,
		Failed
	};

	// Does one step in the AStar search
	State					SearchStep( const Environment &env );



	// Node class + Node management
	struct Node
	{
		Node() {}				// Skip initializing members for performance

		Node*		m_parent;	// The parent of this Node, used in the search
		
		CostType	m_g;		// Cost of this node + all predecessors
		CostType	m_h;		// Heuristic
		CostType	m_f;		// m_g + m_h

		UserType	m_userType;
	};

	struct OpenNode
	{
		OpenNode() {}
		OpenNode(Node* node, CostType f)
			: m_node(node), m_f(f) {}

		Node*		m_node;
		CostType	m_f;
	};

	Node*					AllocateNode();

	void					SetChildPointers( Node *nodeChild );	// Walks up the parent pointers setting the pointers to the children correctly

	// Adds one neighbour to the open list
	void					GetNeighbours( Node *n, const Environment &env );
	bool					AddNeighbour( const UserType &State, CostType added_g_cost );

	// Compare function for the Node heap
	struct HeapNodeCompare
	{
		bool operator() ( const OpenNode& x, const OpenNode& y ) const
		{
			return x.m_f > y.m_f;
		}
	};

	struct OpenNodeEquals
	{
		OpenNodeEquals( const Node* node )
			: m_node(node) {}

		bool operator() ( const OpenNode& other ) const
		{
			return other.m_node == m_node;
		}

		const Node* m_node;
	};

	TouchedList&			m_touchedList;				// Container of open/closed nodes

	typedef std::vector< Node* > NodeVector;
	typedef std::vector< OpenNode > OpenNodeVector;

	OpenNodeVector			m_openList;					// Internally used as a binary heap
	NodeVector				m_neighbours;				// neighbours is a vector filled out by the user each time neighbours to a node are generated

	typedef std::vector< Node* > NodeStartVector;

	NodeStartVector			m_startVector;
	Node *					m_goal;						// Goal of the search

	RegionPool				&m_nodePool;

	typedef std::vector< UserType* > SolutionVector;

	SolutionVector			m_solution;
	typename SolutionVector::reverse_iterator	m_currentSolutionIt;

	Node *					m_lowestHNodeSoFar;			// Used to find an imperfect path when the search fails

	// Debug info:
	unsigned int			m_nodesVisited;
	unsigned int			m_nodesClosed;

	UserType				m_firstOpen;
	bool					m_searchOpen;				// Is this an open search?


	// Disable assignment
	AStar& operator=(const AStar&);
};




/////////////////////////////////////////////////////////////////////
//	Desc.	: Finds a path from start to goal using AStar
//	Param.	: start - start of the path
//			  end   - end of the path
template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
typename AStarSearchResult AStar<UserType,Environment,TouchedList,CostType>::DoFindPath( const UserType &start, const UserType &goal, const Environment &env, unsigned long maxNodes )
{
	return DoFindPath( &start, (&start)+1, goal, env, maxNodes );
}

template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
typename AStarSearchResult AStar<UserType,Environment,TouchedList,CostType>::DoFindPath( const UserType* startsBegin, const UserType* startsEnd, const UserType &goal, const Environment &env, unsigned long maxNodes )
{
	m_openList.reserve(maxNodes / 8);
	m_neighbours.reserve(16);

	// Not an open search
	m_searchOpen = false;

	assert(startsBegin);
	assert(startsEnd);
	assert(startsEnd - startsBegin < 1024);	// Sanity check

	SetStartAndGoal( startsBegin, startsEnd, goal, env );

	size_t numNodes = 0;

	State state = Searching;

	while( (numNodes<maxNodes) && (state == Searching) )
	{
		state = SearchStep( env );
		++numNodes;
	}

	if (state == Succeeded)
	{
		m_lowestHNodeSoFar = NULL;
		return AStarGotPathToGoal;
	}

	// Didn't succeed, check for a best failed node
	if(m_lowestHNodeSoFar)
	{
		SetChildPointers(m_lowestHNodeSoFar);
		return AStarGotPathImperfect;
	}

	// Failed
	FreeAllNodes();

	return AStarNoPath;
}


/////////////////////////////////////////////////////////////////////
//	Desc.	: Finds a path from start to goal using AStar
//	Param.	: start - start of the path
//			  end   - end of the path
//	Author	: nwaanders
//
template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
typename AStarSearchResult AStar<UserType,Environment,TouchedList,CostType>::DoFindOpen( UserType &firstOpen, const UserType &start, const UserType &goal, const Environment &env, unsigned long maxNodes )
{
	m_openList.reserve(maxNodes / 8);
	m_neighbours.reserve(16);

	// An open search (stops whenever the first node was found that meets an exitcondition
	m_searchOpen = true;

	SetStartAndGoal( &start, (&start)+1, goal, env );

	size_t numNodes = 0;

	State state = Searching;

	while( (numNodes < maxNodes) && (state == Searching) )
	{
        state = SearchStep( env );
		++numNodes;
	}

	if (state == Succeeded)
	{		
		firstOpen = m_firstOpen;
		return AStarGotPathToGoal;
	}

	// Failed
	FreeAllNodes();

	return AStarNoPath;
}

template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
void AStar<UserType,Environment,TouchedList,CostType>::SetStartAndGoal( const UserType *startVecBegin, const UserType *startVecEnd, const UserType &goal, const Environment &env )
{	
	assert(startVecBegin);
	assert(startVecEnd);
	assert(startVecEnd - startVecBegin < 1024);	// Sanity check

	FreeAllNodes();	// Remove any memory still allocated by a previous pathfind

	m_nodesVisited = 0;	
    m_nodesClosed = 0;

	// Allocate nodes for start and end
	m_goal	= AllocateNode();

	m_goal->m_g = 0;
	m_goal->m_h = 0;
	m_goal->m_f = 0;
	m_goal->m_parent = 0;
	m_goal->m_userType		= goal;

	m_touchedList.StartNewSearch();

	for(const UserType* startEntry = startVecBegin; startEntry != startVecEnd; ++startEntry)
	{
		Node* startNode = AllocateNode();
		m_startVector.push_back(startNode);
		startNode->m_userType		= *startEntry;
		startNode->m_g = 0;								// Start length is zero
		startNode->m_h = env.Heuristic( *startEntry, goal, 0 );	// Get the heuristic value of the startpoint
		startNode->m_f = startNode->m_g + startNode->m_h;
		startNode->m_parent = 0;							// We don't have a parent
		
		// Push the start node on the Open list
		m_openList.push_back( OpenNode(startNode, startNode->m_f) ); // heap now unsorted
		// Sort back element into heap
		std::push_heap( m_openList.begin(), m_openList.end(), HeapNodeCompare() );

		// Mark it in the touchedList as open
		m_touchedList.Push(*startEntry, startNode, true);
	}

	m_lowestHNodeSoFar = 0;

	m_solution.clear();
	m_currentSolutionIt = m_solution.rend();
}



// Advances search one step
template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
typename AStar<UserType,Environment,TouchedList,CostType>::State AStar<UserType,Environment,TouchedList,CostType>::SearchStep( const Environment &env )
{
	assert(m_goal != 0);	// Make sure we have a goal
	assert(!m_startVector.empty());	// Make sure we have a start

	// If the openlist is empty, stop searching
	if( m_openList.empty() )
	{
		return Failed;
		// Todo: imperfect path, return the path to the point with the lowest f value so far
	}

	// Pop the best node (the one with the lowest f) from the top of the heap
	Node *n = m_openList.front().m_node; // get pointer to the node
	std::pop_heap( m_openList.begin(), m_openList.end(), HeapNodeCompare() );
	m_openList.pop_back();

	// Check for the exit condition, once we pop that we're done
	if( env.TestExitCondition(n->m_userType, m_goal->m_userType) )
	{
		++m_nodesClosed;

		if ( m_searchOpen )
		{
			// This was an open search, so return the firstOpen node
			m_firstOpen = n->m_userType;
		}
		else
		{
			// This wasn't an open search, so return the full path to the end node

			// Set the child pointers in each node (except Goal which has no child)
			SetChildPointers(n);
		}

		return Succeeded;
	}
	else
	{
		// Add the neighbours of the current node to the openlist
		m_neighbours.clear();

		GetNeighbours(n, env);

		NodeVector::iterator i = m_neighbours.begin();
		NodeVector::iterator e = m_neighbours.end();

		for (; i!=e; ++i)
		{
			++m_nodesVisited;	

			Node *neighbour_node = *i;

			neighbour_node->m_g += n->m_g;	// Add the g cost of the current node to the neighbour node gcost. This makes the m_g variable the total cost to get in the neighbour node

			// Now we need to find whether the node is on the open or closed lists
			// If it is on the openlist but the node that is already on them is better (lower g)
			// then we can forget about this successor


			// Try to find the node in the open list
			TouchedList::const_iterator touchedIterator = m_touchedList.Find( neighbour_node->m_userType );

			bool bIsTouched = touchedIterator != m_touchedList.End();
			bool bIsOpen = bIsTouched && touchedIterator->m_bOpen;

			if( bIsOpen )
			{
				// Node is on the open list

				Node* open_node = static_cast<Node*>( m_touchedList.GetNode( touchedIterator ) );

				if( neighbour_node->m_g < open_node->m_g )
				{
					// The new neighbour is cheaper than the one in the open list.  Update the open list node.

					// Copy valid data from neighbour_node to open_node (only g value, since usertype matches)
					const float new_g = neighbour_node->m_g;
					open_node->m_g = new_g;

					// The change in g has a *very* minor effect on the h, so don't bother updating it from the old value.
					// open_node->m_h = env.Heuristic( open_node->m_userType, m_goal->m_userType, new_g );

					open_node->m_f = new_g + open_node->m_h;
					open_node->m_parent = n;

					// Find the open node in the heap and update the heap
					OpenNodeVector::iterator openlist_result = std::find_if( m_openList.begin(), m_openList.end(), OpenNodeEquals(open_node) );
					assert( openlist_result != m_openList.end() );
					openlist_result->m_f = open_node->m_f;
					std::push_heap( m_openList.begin(), openlist_result + 1, HeapNodeCompare() );

				#ifdef _DEBUG
					// If your STL has std::is_heap, this could be handy
					//assert(std::is_heap(m_openList.begin(), m_openList.end(), HeapNodeCompare() ));
				#endif
				}

				// Either the new neighbour was updated into the heap or was more expensive (thus ignored)
				continue;
			}
			else	// Item isn't open, so it's either closed or untouched
			{
				if( bIsTouched )
				{
					// It's been closed before

					assert( !bIsOpen );

					Node* closed_node = static_cast<Node*>( m_touchedList.GetNode( touchedIterator ) );

					// we found this neighbour on the closedlist
					if( closed_node->m_g <= neighbour_node->m_g )
					{
						// The one on the closedList is cheaper than this one - ignore it, and it will be freed by the clear region at the end
						continue;
					}
					else
					{
						// The new neighbor node is cheaper than the one in the closed list.  Update the closed list node.

						// Copy valid data from neighbour_node to closed_node (only g value, since usertype matches)
						closed_node->m_g = neighbour_node->m_g;

						// Throw away the neighbour_node and use the old closed_node in its place.  The neighbour is freed by the clear at the end.
						neighbour_node = closed_node;

						// The touched list entry will be updated below
					}
				}

				// This node is the best node so far, so we need to add it to the openlist

				// Update the node values
				neighbour_node->m_parent = n;
				neighbour_node->m_h = env.Heuristic( neighbour_node->m_userType, m_goal->m_userType, neighbour_node->m_g ); // OPTIMIZATION: You can probably skip this if !bIsTouched, since it should be already set
				neighbour_node->m_f = neighbour_node->m_g + neighbour_node->m_h;

				// Put it in the touched list as open
				m_touchedList.Push( neighbour_node->m_userType, neighbour_node, true );

				// Push it onto the openlist priority q
				m_openList.push_back( OpenNode(neighbour_node, neighbour_node->m_f) );											// heap now unsorted
				std::push_heap( m_openList.begin(), m_openList.end(), HeapNodeCompare() );		// sort back element into heap
			}
		}

		// Mark n closed on the touchedList, as we have expanded it now
		++m_nodesClosed;
		TouchedList::iterator nIt = m_touchedList.Find(n->m_userType);

#ifdef _DEBUG
		assert(nIt != m_touchedList.End());
		assert(m_touchedList.GetNode(nIt) == n);
#endif

		m_touchedList.Update(nIt, false);

		// If this is the nodew with the lowest H value we've seen so far, remember it.
		if((!m_lowestHNodeSoFar) || (m_lowestHNodeSoFar->m_h > n->m_h))
		{
			if(env.ValidatePartialResult( n->m_userType ))
			{
				m_lowestHNodeSoFar = n;
			}
		}
	}

 	return Searching;
}



template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
void AStar<UserType,Environment,TouchedList,CostType>::GetNeighbours( Node *n, const Environment &env )
{
	static std::vector< UserType > neighbours;	// Don't reallocate every time this is called
	static std::vector< CostType > gcost;

	neighbours.clear();
	gcost.clear();

	if (n->m_parent)
	{
		env.GetNeighbours( n->m_parent->m_userType, n->m_userType, neighbours, gcost );	// Pass the parent in and the current point (This way we don't have to add all neighbours)
	}
	else
	{
		env.GetNeighbours( n->m_userType, n->m_userType, neighbours, gcost );	// Pass the same point in twice. 
	}

	assert(neighbours.size() == gcost.size());	// If it stops here, you didn't provide the same number of neighbours and g-costs. They should be the same (one gcost for every neighbour)

	std::vector< UserType >::iterator i = neighbours.begin();
	std::vector< UserType >::iterator e = neighbours.end();
	std::vector< CostType >::iterator g_i = gcost.begin();

	for (; i!=e; ++i, ++g_i)
	{
        AddNeighbour(*i, *g_i);
	}
}



// User calls this to add a successor to a list of neighbours
// when expanding the search frontier
template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
bool AStar<UserType,Environment,TouchedList,CostType>::AddNeighbour( const UserType &state, CostType added_g_cost )
{
	Node *node = AllocateNode();

	if( node )
	{
		node->m_userType = state;
		node->m_g = added_g_cost;

		m_neighbours.push_back( node );

		return true;
	}

	return false;
}



// Node Management:

// This is called when a search fails or is cancelled to free all used memory 
template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
void AStar<UserType,Environment,TouchedList,CostType>::FreeAllNodes()
{
	m_openList.clear();

	m_startVector.clear();	// Clear out the pointer to a new deleted node
	m_goal = 0;

	m_lowestHNodeSoFar = 0;

	m_solution.clear();
	m_currentSolutionIt = m_solution.rend();

	// Clear the memory pool:
	m_nodePool.Clear();
}


// Node memory management
template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
typename AStar<UserType,Environment,TouchedList,CostType>::Node *AStar<UserType,Environment,TouchedList,CostType>::AllocateNode()
{
	return static_cast<Node*>(m_nodePool.Allocate(sizeof Node));
}


template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
void AStar<UserType,Environment,TouchedList,CostType>::SetChildPointers( Node *nodeChild )
{
	assert(nodeChild);
	assert(m_solution.empty());

	while( nodeChild )
	{
		m_solution.push_back(&nodeChild->m_userType);
		nodeChild = nodeChild->m_parent;
	}

	m_currentSolutionIt = m_solution.rbegin();
}

// Functions for traversing the solution

// Get start node
template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
UserType *AStar<UserType,Environment,TouchedList,CostType>::GetSolutionStart()
{
	m_currentSolutionIt = m_solution.rbegin();
	if(m_currentSolutionIt == m_solution.rend())
	{
		return NULL;
	}
	else
	{
		return *m_currentSolutionIt;
	}
}



// Get next node
template <typename UserType, typename Environment, typename TouchedList, typename CostType> 
UserType *AStar<UserType,Environment,TouchedList,CostType>::GetSolutionNext()
{
	if(m_currentSolutionIt != m_solution.rend())
	{
		if(++m_currentSolutionIt != m_solution.rend())
		{
			return *m_currentSolutionIt;
		}
	}

	return NULL;
}

#endif