// (C) 2009 THQ Inc. Not for commercial use or redistribution without permission. All rights reserved.

#pragma once

#ifndef PATHPRECISE_UTIL_H
#define PATHPRECISE_UTIL_H

#include "TerrainMap.h"

#pragma warning( disable : 4127 ) // warning C4127: conditional expression is constant

namespace PPUtil
{
	// Change this if you want to make the likelyhood of big squads moving through small passages.
	const float HIGHWAY_SIZE_WEIGHT = 0.1f;			// This times (highwaysize - passable size) is added to the cost for a node that doesn't meet highwaysize reqs.
	const float PRIORITY_WEIGHT = 0.0f;				// Weight added to a positition that is blocked by a lower priority entity (priority difference is ignored)
	const float REF_WEIGHT = 0.4f;					// This * the cost to get to the ref cell is added to find open calls
	//const float GVAL_TIEBREAKER_SCALE = 0.0f;	// This * gval so far is subtracted from the path cost to break ties toward nodes farther from the start
	const float H_TIEBREAKER_SCALE = 1.0001f;

	const float DISTANCE_TO_COST_SCALE = 1.0f;													// Distance times this is the basis for cost
	const float DIAGONAL_COST = DISTANCE_TO_COST_SCALE * 1.4142135623730950488016887242097f;	// Cost to move diagonal
	const float CARDINAL_COST = DISTANCE_TO_COST_SCALE;											// Cost to move N/S/E/W
	const float DIAGONAL_COST_MINUS_CARDINAL = DIAGONAL_COST - CARDINAL_COST;					// Cost to move diagonal less cost to move cardinal

	//
	// Helper function for adding neighbours. 
	// Uses bitmasks to filter off all neighbours that are illegal. Then returns the neighbours and their gcosts.
	//
	template < typename CostType, bool checkCameFrom, bool checkPassible, bool checkDiagonals, bool checkHighway, bool checkPriority, bool checkPreciseOnly, bool ignoreMoving >
	inline void AddNeighbours( const Point2i &parent_point, const Point2i &point, std::vector< Point2i > &neighbours, std::vector< CostType > &gcost, const TerrainMap* terrainMap )
	{
		// If the point is not inside the bounds, return immediately
		if ((point.x<0) || (point.x >= terrainMap->GetWidth()) ||
			(point.y<0) || (point.y >= terrainMap->GetHeight()))
		{
			return;
		}

		//
		// The next piece of code uses bitmasks to determine what the valid neighbours are going to be.
		// The bits are setup as follows:
		//
		//	+---+---+---+
		//	| 7 | 0 | 1 |	bit nr's in the bitmask
		//	+---+---+---+
		//	| 6 |   | 2 |
		//	+---+---+---+
		//	| 5 | 4 | 3 |
		//	+---+---+---+
		//

		enum Directions
		{
			dir_none	= 0x00,
			dir_n		= 0x01,
			dir_ne		= 0x02,
			dir_e		= 0x04,
			dir_se		= 0x08,
			dir_s		= 0x10,
			dir_sw		= 0x20,
			dir_w		= 0x40,
			dir_nw		= 0x80,
			dir_all		= 0xFF,
		};
		enum OrderOfDirection
		{
			order_none = 0,
			order_n,
			order_ne,
			order_e,
			order_se,
			order_s,
			order_sw,
			order_w,
			order_nw,
			order_numDirections
		};

		// By default, tag all directions:
		unsigned char bitmask = dir_all;



		// Filter off invalid positions because of worldbounds:

		// These masks describe the possible directions left if the current position is bound by that direction. 
		static const unsigned char westBoundMask  = dir_n | dir_ne | dir_e | dir_se | dir_s;
		static const unsigned char eastBoundMask  = dir_s | dir_sw | dir_w | dir_nw | dir_n;
		static const unsigned char northBoundMask = dir_e | dir_se | dir_s | dir_sw | dir_w;
		static const unsigned char southBoundMask = dir_w | dir_nw | dir_n | dir_ne | dir_e;

		// We already know that the point is inside the map, but it might be at the border:
		// Filter off all the directions that we know we can't go to:
		if (point.x == 0)							bitmask &= westBoundMask;
		if (point.x == terrainMap->GetWidth()-1)	bitmask &= eastBoundMask;
		if (point.y == 0)							bitmask &= southBoundMask;
		if (point.y == terrainMap->GetHeight()-1)	bitmask &= northBoundMask;

		if (!checkDiagonals)
		{
			// Filter off the diagonals:
			bitmask &= dir_n | dir_e | dir_s | dir_w;
		}

		// Now filter off the directions that we already should have visited by looking at where our parent is
		// 
		// Possible directions (x) when coming from the parent P, and currently in point S
		//
		//	+---+---+---+
		//	|   | x | x |	
		//	+---+---+---+
		//	| P | S | x |		So there's 5 new neighbours to add if we came from the north, south, east or west 
		//	+---+---+---+		(we're adding north and south as well in this case, because we might have cut a diagonal, and therefore we didn't check those cells)
		//	|   | x | x |
		//	+---+---+---+
		// 
		//	+---+---+---+
		//	| P |   | x |	
		//	+---+---+---+	
		//	|   | S | x |		And there's 5 neighbours to add if we came from the nw, ne, sw or se
		//	+---+---+---+
		//	| x | x | x |
		//	+---+---+---+
		// 
		//

		if (checkCameFrom)
		{
			static const unsigned char cameFromMask[]	= {
				dir_all,									// none	0
				dir_sw | dir_s | dir_se | dir_w | dir_e,	// n	1
				dir_nw | dir_w | dir_sw | dir_s | dir_se,	// ne	2
				dir_nw | dir_w | dir_sw | dir_n | dir_s,	// e	3
				dir_sw | dir_w | dir_nw | dir_n | dir_ne,	// se	4
				dir_nw | dir_n | dir_ne | dir_w | dir_e,	// s	5
				dir_nw | dir_n | dir_ne | dir_e | dir_se,	// sw	6
				dir_ne | dir_e | dir_se | dir_n | dir_s,	// w	7
				dir_ne | dir_e | dir_se | dir_s | dir_sw,	// nw	8
			};

			Point2i parent_delta = point - parent_point;

			if (parent_delta.x < -1 || 
				parent_delta.x > 1 ||
				parent_delta.y < -1 || 
				parent_delta.y > 1)
			{
				// Bigger jump than just one tile, so just leave all directions open
				parent_delta.x = 0;	// 0,0 is adding all neighbours.
				parent_delta.y = 0;
			}

			// Make the parent_delta go from 0 to 2, instead of from -1 to 1
			++parent_delta.x;
			++parent_delta.y;

			//		   	  		                   y  x	
			static const int convertDirectionTable[3][3] = {
				{ order_ne, order_n,	order_nw },	// { dir_ne,	dir_n,		dir_nw },
				{ order_e,	order_none, order_w },	// { dir_e,		dir_none,	dir_w, },
				{ order_se, order_s,	order_sw },	// { dir_se,	dir_s,		dir_sw },
			};

			// Filter off all the directions that we can't go to from this position
			bitmask &= cameFromMask[ convertDirectionTable[parent_delta.y][parent_delta.x] ];
		}




		// Filter off any positions that are impassible. Also, at the same time, filter off any diagonals that we can't reach.
		// A diagonal can't be reached if the n,s,e,w directions are blocked. If we would go to the diagonal, we would cut the corner, and we don't want that

		// passiblePosition is a flag that checks the passiblePositions around the current position.
		// If a certain direction is free, it checks that direction to see if it's passible. If not, it filters off that direction.
		// At the same time it filters off all diagonals if the primary directions (n, s, e, w) are blocked

		if (checkPassible)
		{
			struct PassStruct
			{
				Point2i        m_delta;
				unsigned char		m_bitmask;
				OrderOfDirection	m_order;
			};

			static const PassStruct passiblePosition[]	=	{
				{  Point2i( 0,  1),  dir_ne | dir_n | dir_nw, order_n },		// Check the N, S, E, W directions first, so that we filter off diagonals asap
				{  Point2i( 0, -1),  dir_se | dir_s | dir_sw, order_s },	
				{  Point2i( 1,  0),  dir_ne | dir_e | dir_se, order_e },
				{  Point2i(-1,  0),  dir_nw | dir_w | dir_sw, order_w },
				{  Point2i( 1,  1),  dir_ne,				   order_ne },
				{  Point2i( 1, -1),  dir_se,				   order_se },
				{  Point2i(-1,  1),  dir_nw,				   order_nw },
				{  Point2i(-1, -1),  dir_sw,				   order_sw },
			};

			for (int pass=0; pass<8; ++pass)
			{
				// Only check the directions that are still left in our bitmask
				if ( (bitmask & passiblePosition[pass].m_bitmask) != 0 )
				{
					if(!terrainMap->IsPassable(point + passiblePosition[pass].m_delta))
					{
						bitmask &= ~passiblePosition[pass].m_bitmask;
					}
				}
			}
		}

		// Now add all the remaining neighbours
		struct Dx
		{
			int					xoffset; 
			int					yoffset;
			CostType			distance;
			OrderOfDirection	m_order;
		};

		static const Dx		deltas[] = {	
			{ 0, +1, CARDINAL_COST,	order_n  },	// n
			{+1, +1, DIAGONAL_COST,	order_ne },	// ne 
			{+1,  0, CARDINAL_COST,	order_e  },	// e
			{+1, -1, DIAGONAL_COST, order_se },	// se 
			{ 0, -1, CARDINAL_COST, order_s  },	// s
			{-1, -1, DIAGONAL_COST, order_sw },	// sw 
			{-1,  0, CARDINAL_COST, order_w  },	// w
			{-1, +1, DIAGONAL_COST, order_nw },	// nw 
		};

		// All the impossible neighbours have now been filtered off
		// Go through the bitmask and add all neighbours who's bit was set:
		for (int i=0; i<8; ++i)
		{
			if ((bitmask & (1<<i)) != 0)
			{
				const Dx& delta = deltas[i];

				Point2i pt = point + Point2i(delta.xoffset, delta.yoffset);
				// Add the neighbour
				neighbours.push_back( pt );

				CostType g = delta.distance;

				// the gcost might 
				gcost.push_back( g );
			}
		}
	}
}

#endif